==> fwobac-2.1.tar.gz: fwobac-2.1/conf_parse.c <==
/* $Id: conf_parse.c,v 1.5 2020/02/07 00:28:02 rlr Exp $ */

/*
 * Copyright (c) 2017, 2018, 2020 Ross Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/mman.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "conf_parse.h"


/*
 * Finite State Machine for conf_parse_buf()
 *
 * ========
 * State
 *     input	action	==> new state
 * zz is the end of buffer
 * NUL is '\0'
 * TAB is '\t'
 * NL is '\n'
 * SPACE is '\t'
 * # is '#'
 * ch is any other character
 * ========
 *
 * CS_SOL
 *     zz	-	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> CS_ERR
 *     NL	-	== -
 *     #	-	==> CS_CMT
 *     ch	-	==> CS_NAME
 *
 * CS_CMT
 *     zz	-	==> CS_END
 *     NUL	-	==> CS_ERR
 *     NL	-	==> CS_SOL
 *     ch	-	==> -
 *
 * CS_NAME
 *     zz	cb	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> CS_NVS
 *     NL	cb	==> CS_SOL
 *     ch	-	==> -
 *
 * CS_NVS
 *     zz	cb	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> -
 *     NL	cb	==> CS_SOL
 *     ch	-	==> CS_VAL
 *
 * CS_VAL
 *     zz	cb	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> -
 *     NL	cb	==> CS_SOL
 *     ch	-	==> -
 *
 * CS_END
 *
 * CS_ERR
 *
 */

enum fsm_state {
	CS_SOL,   /* start of line */
	CS_CMT,   /* in a comment */
	CS_NAME,  /* in name */
	CS_NVS,   /* in name/value separating whitespace */
	CS_VAL,   /* value */
	CS_END,
	CS_ERR
};


int
conf_parse_buf(const char *buf, size_t len,
    int (*cb)(const char *, size_t, const char *, size_t, void *), void *a)
{
	int	 result;

	result = CP_NO_ERR;

	if (len > 0) {
		size_t	 i, n0, nl, v0, line;
		enum fsm_state	 st;

		n0 = 0; nl = 0; v0 = 0;  /* Avoid compiler warnings. */

		/* Implement the FSM. */
		line = 1;
		i = 0;
		st = CS_SOL;
		while (st != CS_END && st != CS_ERR) {
			switch (st) {

			case CS_SOL :
				if (i == len) {
					st = CS_END;
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					st = CS_ERR;
				} else if (buf[i] == '\n') {
					line++;
					/* no change */
				} else if (buf[i] == '#') {
					st = CS_CMT;
				} else {
					n0 = i;
					st = CS_NAME;
				}
			break;

			case CS_CMT :
				if (i == len) {
					st = CS_END;
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\n') {
					line++;
					st = CS_SOL;
				} else {
					/* no change */
				}
			break;

			case CS_NAME :
				if (i == len) {
					if (cb(&buf[n0], i - n0, NULL, 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						st = CS_END;
					}
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					nl = i - n0;
					st = CS_NVS;
				} else if (buf[i] == '\n') {
					if (cb(&buf[n0], i - n0, NULL, 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						line++;
						st = CS_SOL;
					}
				} else {
					/* no change */
				}
			break;

			case CS_NVS :
				if (i == len) {
					if (cb(&buf[n0], nl, &buf[i - 1], 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						st = CS_END;
					}
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					/* no change */
				} else if (buf[i] == '\n') {
					if (cb(&buf[n0], nl, &buf[i], 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						line++;
						st = CS_SOL;
					}
				} else {
					v0 = i;
					st = CS_VAL;
				}
			break;

			case CS_VAL :
				if (i == len) {
					if (cb(&buf[n0], nl, &buf[v0], i - v0,
					    a) != 0) {
						st = CS_ERR;
					} else {
						st = CS_END;
					}
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					/* no change */
				} else if (buf[i] == '\n') {
					if (cb(&buf[n0], nl, &buf[v0], i - v0,
					    a) != 0) {
						st = CS_ERR;
					} else {
						line++;
						st = CS_SOL;
					}
				} else {
					/* no change */
				}
			break;

			default :
				/* Internal error occurred. */
				exit(1);
			break;

			}

			i++;
		}

		if (st == CS_ERR) {
			result = line;
		}
	}

	return result;
}


int
conf_parse_file(const char *filepath,
    int (*cb)(const char *, size_t, const char *, size_t, void *), void *a)
{
	struct	stat	 st;
	unsigned char	*d;
	int	 fd;
	int	 result;

	if ((fd = openat(AT_FDCWD, filepath, O_RDONLY, 0)) == -1) {
		result = CP_OPEN_ERR;
	} else if (fstat(fd, &st) == -1) {
		result = CP_STAT_ERR;
	} else if (st.st_size > SSIZE_MAX) {
		result = CP_TOO_BIG_ERR;
	} else if (st.st_size <= 0) {
		/* File is empty so there is nothign to do. */
		result = CP_NO_ERR;
	} else if ((d = mmap(NULL, (size_t)st.st_size, PROT_READ, MAP_SHARED,
	    fd, 0)) == MAP_FAILED) {
		result = CP_MMAP_ERR;
	} else {
		result = conf_parse_buf((char *)d, (size_t)st.st_size, cb, a);
		munmap(d, (size_t)st.st_size);
	}

	if (fd != -1) {
		close(fd);
	}

	return result;
}

==> fwobac-2.3.tar.gz: fwobac-2.3/conf_parse.c <==
/* $Id: conf_parse.c,v 1.6 2023/11/19 05:26:45 rlr Exp $ */

/*
 * Copyright (c) 2017, 2018, 2020, 2023 Ross Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/mman.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "conf_parse.h"


/*
 * Finite State Machine for conf_parse_buf()
 *
 * ========
 * State
 *     input	action	==> new state
 * zz is the end of buffer
 * NUL is '\0'
 * TAB is '\t'
 * NL is '\n'
 * SPACE is '\t'
 * # is '#'
 * ch is any other character
 * ========
 *
 * CS_SOL
 *     zz	-	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> CS_ERR
 *     NL	-	== -
 *     #	-	==> CS_CMT
 *     ch	-	==> CS_NAME
 *
 * CS_CMT
 *     zz	-	==> CS_END
 *     NUL	-	==> CS_ERR
 *     NL	-	==> CS_SOL
 *     ch	-	==> -
 *
 * CS_NAME
 *     zz	cb	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> CS_NVS
 *     NL	cb	==> CS_SOL
 *     ch	-	==> -
 *
 * CS_NVS
 *     zz	cb	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> -
 *     NL	cb	==> CS_SOL
 *     ch	-	==> CS_VAL
 *
 * CS_VAL
 *     zz	cb	==> CS_END
 *     NUL	-	==> CS_ERR
 *     TAB,SP	-	==> -
 *     NL	cb	==> CS_SOL
 *     ch	-	==> -
 *
 * CS_END
 *
 * CS_ERR
 *
 */

enum fsm_state {
	CS_SOL,   /* start of line */
	CS_CMT,   /* in a comment */
	CS_NAME,  /* in name */
	CS_NVS,   /* in name/value separating whitespace */
	CS_VAL,   /* value */
	CS_END,
	CS_ERR
};


int
conf_parse_buf(const char *buf, size_t len,
    int (*cb)(const char *, size_t, const char *, size_t, void *), void *a)
{
	int	 result;

	result = CP_NO_ERR;

	if (len > 0) {
		size_t	 i, n0, nl, v0, line;
		enum fsm_state	 st;

		n0 = 0; nl = 0; v0 = 0;  /* Avoid compiler warnings. */

		/* Implement the FSM. */
		line = 1;
		i = 0;
		st = CS_SOL;
		while (st != CS_END && st != CS_ERR) {
			switch (st) {

			case CS_SOL :
				if (i == len) {
					st = CS_END;
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					st = CS_ERR;
				} else if (buf[i] == '\n') {
					line++;
					/* no change */
				} else if (buf[i] == '#') {
					st = CS_CMT;
				} else {
					n0 = i;
					st = CS_NAME;
				}
			break;

			case CS_CMT :
				if (i == len) {
					st = CS_END;
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\n') {
					line++;
					st = CS_SOL;
				} else {
					/* no change */
				}
			break;

			case CS_NAME :
				if (i == len) {
					if (cb(&buf[n0], i - n0, NULL, 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						st = CS_END;
					}
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					nl = i - n0;
					st = CS_NVS;
				} else if (buf[i] == '\n') {
					if (cb(&buf[n0], i - n0, NULL, 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						line++;
						st = CS_SOL;
					}
				} else {
					/* no change */
				}
			break;

			case CS_NVS :
				if (i == len) {
					if (cb(&buf[n0], nl, &buf[i - 1], 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						st = CS_END;
					}
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					/* no change */
				} else if (buf[i] == '\n') {
					if (cb(&buf[n0], nl, &buf[i], 0, a)
					    != 0) {
						st = CS_ERR;
					} else {
						line++;
						st = CS_SOL;
					}
				} else {
					v0 = i;
					st = CS_VAL;
				}
			break;

			case CS_VAL :
				if (i == len) {
					if (cb(&buf[n0], nl, &buf[v0], i - v0,
					    a) != 0) {
						st = CS_ERR;
					} else {
						st = CS_END;
					}
				} else if (buf[i] == '\0') {
					st = CS_ERR;
				} else if (buf[i] == '\t' || buf[i] == ' ') {
					/* no change */
				} else if (buf[i] == '\n') {
					if (cb(&buf[n0], nl, &buf[v0], i - v0,
					    a) != 0) {
						st = CS_ERR;
					} else {
						line++;
						st = CS_SOL;
					}
				} else {
					/* no change */
				}
			break;

			default :
				/* Internal error occurred. */
				exit(1);
			break;

			}

			i++;
		}

		if (st == CS_ERR) {
			result = line;
		}
	}

	return result;
}


int
conf_parse_file(const char *filepath,
    int (*cb)(const char *, size_t, const char *, size_t, void *), void *a)
{
	struct	stat	 st;
	unsigned char	*d;
	int	 fd;
	int	 result;

	if ((fd = openat(AT_FDCWD, filepath, O_RDONLY, 0)) == -1) {
		result = CP_OPEN_ERR;
	} else if (fstat(fd, &st) == -1) {
		result = CP_STAT_ERR;
	} else if (st.st_size > SSIZE_MAX) {
		result = CP_TOO_BIG_ERR;
	} else if (st.st_size <= 0) {
		/* File is empty so there is nothing to do. */
		result = CP_NO_ERR;
	} else if ((d = mmap(NULL, (size_t)st.st_size, PROT_READ, MAP_SHARED,
	    fd, 0)) == MAP_FAILED) {
		result = CP_MMAP_ERR;
	} else {
		result = conf_parse_buf((char *)d, (size_t)st.st_size, cb, a);
		munmap(d, (size_t)st.st_size);
	}

	if (fd != -1) {
		close(fd);
	}

	return result;
}
